/*
 * Pound - the reverse-proxy load-balancer
 * Copyright (C) 2002-2020 Apsis GmbH
 *
 * This file is part of Pound.
 *
 * Pound is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pound is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/> .
 *
 * Contact information:
 * Apsis GmbH
 * P.O.Box
 * 8707 Uetikon am See
 * Switzerland
 * EMail: roseg@apsis.ch
 */

#define HAVE_MANDATORY_INCLUDES
#define HAVE_PCRE2POSIX_H
/* #undef HAVE_PCRE_PCRE2POSIX_H */
#define HAVE_PCREPOSIX_H
/* #undef HAVE_PCRE_PCREPOSIX_H */
#define HAVE_REGEX_H
#define HAVE_STDARG_H

#define VERSION            "3.0.2"
#define MAXBUF             8192
#define MAXHEADERS         128
#define RESURRECT_TO       15
#define RESURRECT_CYCLE    60
#define HASH_FUNCTION      HASH_FNV
#define LOG_FACILITY       LOG_DAEMON
#define H2TABSIZE          4096
#define H2FRAMESIZE        16384

#ifdef  HAVE_MANDATORY_INCLUDES
#include    <stdio.h>
#include    <pthread.h>
#include    <yaml.h>
#include    <nanomsg/nn.h>
#include    <nanomsg/inproc.h>
#include    <nanomsg/pipeline.h>
#include    <nanomsg/pair.h>
#include    <nanomsg/reqrep.h>
#include    <stdlib.h>
#include    <unistd.h>
#include    <fcntl.h>
#include    <ctype.h>
#include    <getopt.h>
#include    <string.h>
#include    <syslog.h>
#include    <sys/types.h>
#include    <sys/socket.h>
#include    <netdb.h>
#include    <sys/stat.h>
#include    <time.h>
#include    <poll.h>
#include    <semaphore.h>
#include    <pwd.h>
#include    <grp.h>
#include    <signal.h>
#include    <setjmp.h>
#include    <mbedtls/build_info.h>
#include    <mbedtls/oid.h>
#include    <mbedtls/asn1.h>
#include    <mbedtls/x509.h>
#include    <mbedtls/entropy.h>
#include    <mbedtls/ctr_drbg.h>
#include    <mbedtls/net_sockets.h>
#include    <mbedtls/ssl.h>
#include    <mbedtls/error.h>
#endif

#ifdef  HAVE_STDARG_H
#include    <stdarg.h>
#else
#include    <varargs.h>
#endif

#ifdef  HAVE_PCRE2POSIX_H
#include    <pcre2posix.h>
#else
#ifdef HAVE_PCRE_PCRE2POSIX_H
#include    <pcre/pcre2posix.h>
#else
#ifdef  HAVE_PCREPOSIX_H
#include    <pcreposix.h>
#else
#ifdef HAVE_PCRE_PCREPOSIX_H
#include    <pcre/pcreposix.h>
#else
#ifdef HAVE_REGEX_H
#include    <regex.h>
#endif
#endif
#endif
#endif
#endif


#include    "uthash.h"
#include    "utarray.h"
#include    "hpack.h"

#define F_CONF      "/etc/pound/pound.yaml"
#define F_PID       "/var/run/pound.pid"

#ifndef NI_MAXHOST
#define NI_MAXHOST  1025
#endif

#ifndef NI_MAXSERV
#define NI_MAXSERV  32
#endif

/* Data structures */

typedef struct {
    char    *root_jail;
    uid_t   user;
    gid_t   group;
    char    *pid;
    char    *err404;
    char    *err405;
    char    *err500;
    char    *http2_preamble[5];
    int     log_facility;
    int     log_level;
} GLOBAL;

extern GLOBAL   global;

typedef struct _backend_ {
    yaml_node_item_t    id;
    struct addrinfo *addr;
    /* int          rewriteLocation; */
    /* int          rewriteDestination; */
    char            *add_header;
    int             alive;
    int             timeout;
    int             threads;
    int             is_dead;
    int             sock_in;
    int             sock;
} BACKEND;

extern BACKEND  *backends;
extern int      backends_len;

typedef struct _service_ {
    regex_t             *url;
    int                 session;
    regex_t             *head_require;
    regex_t             *head_deny;
    BACKEND             **backends;
    int                 backends_len;
    int                 sock_in;
    int                 sock;
} SERVICE;

typedef struct {
    mbedtls_x509_crt    certificate;
    mbedtls_pk_context  key;
    regex_t             *host;
    int                 host_len;
} SNI;

typedef struct _http_listener_ {
    struct addrinfo         *addr;
    int                     client;
    int                     threads;
    SERVICE                 **services;
    int                     services_len;
    mbedtls_ssl_config      *conf;
    SNI                     **sni;
    int                     sock_in;
    int                     sock_fan;
} HTTP_LISTENER;

extern HTTP_LISTENER    *http_listeners;
extern int              http_len;
extern mbedtls_ctr_drbg_context tls_ctr_drbg;

extern regex_t rex_Response, rex_Expect, rex_ContentLength, rex_Chunked, rex_Connection_Closed, rex_Connection_HTTP2, rex_Upgrade_HTTP2;

extern sem_t    sem_start;

/* config.c */
extern void config(const int, char **);

/* util.c */
extern void logmsg(int, const char *, ...);
extern void *thr_resurrect(void *);
extern void time_stamp(char *);
extern int  do_sni(void *, mbedtls_ssl_context *, const unsigned char *, size_t);

/* backend.c */
extern void *thr_backend(void *);

/* http.c */
extern void *thr_service(void *);
extern int get_be(HTTP_LISTENER *, char *, char *, char *[], struct hpack_headerblock *);
extern void *thr_http(void *);

/* http2.c */
extern void do_http2(HTTP_LISTENER *, FILE *, char *, char *, int);
